@@ -1,9 +1,9 @@
 /*
- *	$Id: lspci.c,v 1.3 1997/12/27 11:57:12 mj Exp $
+ *	$Id: lspci.c,v 1.6 1998/02/08 10:58:54 mj Exp $
  *
  *	Linux PCI Utilities -- List All PCI Devices
  *
- *	Copyright (c) 1997 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
+ *	Copyright (c) 1997, 1998 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -28,8 +28,9 @@ static int func_filter = -1;
 static int vend_filter = -1;
 static int dev_filter = -1;
 static int show_tree;			/* Show bus tree */
+static char *pci_dir = PROC_BUS_PCI;
 
-static char options[] = "nvbxB:S:F:V:D:t";
+static char options[] = "nvbxB:S:F:V:D:ti:p:";
 
 static char help_msg[] = "\
 Usage: lspci [<switches>]\n\
@@ -40,11 +41,13 @@ Usage: lspci [<switches>]\n\
 -x\tShow hex-dump of config space (-xx shows full 256 bytes)\n\
 -B <bus>, -S <slot>, -F <func>, -V <vendor>, -D <device>  Show only selected devices\n\
 -t\tShow bus tree\n\
+-i <file>\tUse specified ID database instead of " ETC_PCI_IDS "\n\
+-p <dir>\tUse specified bus directory instead of " PROC_BUS_PCI "\n\
 ";
 
 /* Format strings used for IRQ numbers */
 
-#ifdef __sparc_v9__
+#ifdef ARCH_SPARC64
 #define IRQ_FORMAT "%08x"
 #else
 #define IRQ_FORMAT "%d"
@@ -96,10 +99,12 @@ scan_dev_list(void)
 {
   FILE *f;
   byte line[256];
+  byte name[256];
 
-  if (! (f = fopen(PROC_BUS_PCI "/devices", "r")))
+  sprintf(name, "%s/devices", pci_dir);
+  if (! (f = fopen(name, "r")))
     {
-      perror("Unable to open " PROC_BUS_PCI "/devices");
+      perror(name);
       exit(1);
     }
   while (fgets(line, sizeof(line), f))
@@ -134,8 +139,8 @@ scan_dev_list(void)
 static inline void
 make_proc_pci_name(struct device *d, char *p)
 {
-  sprintf(p, PROC_BUS_PCI "/%02x/%02x.%x",
-	  d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
+  sprintf(p, "%s/%02x/%02x.%x",
+	  pci_dir, d->bus, PCI_SLOT(d->devfn), PCI_FUNC(d->devfn));
 }
 
 static void
@@ -143,7 +148,7 @@ scan_config(void)
 {
   struct device *d;
   char name[64];
-  int fd;
+  int fd, res;
   int how_much = (show_hex > 1) ? 256 : 64;
 
   for(d=first_dev; d; d=d->next)
@@ -154,11 +159,17 @@ scan_config(void)
 	  fprintf(stderr, "lspci: Unable to open %s: %m\n", name);
 	  exit(1);
 	}
-      if (read(fd, d->config, how_much) != how_much)
+      res = read(fd, d->config, how_much);
+      if (res < 0)
 	{
 	  fprintf(stderr, "lspci: Error reading %s: %m\n", name);
 	  exit(1);
 	}
+      if (res != how_much)
+	{
+	  fprintf(stderr, "lspci: Only %d bytes of config space available to you\n", res);
+	  exit(1);
+	}
       close(fd);
     }
 }
@@ -273,18 +284,24 @@ show_bases(struct device *d, int cnt)
 	pos = d->kernel_base_addr[i];
       if (!pos || pos == 0xffffffff)
 	continue;
-      if (verbose > 1)
-	printf("\tRegion %d: ", i);
-      else
-	putchar('\t');
       if (flg & PCI_BASE_ADDRESS_SPACE_IO)
 	{
 	  if (cmd & PCI_COMMAND_IO)
-	    printf("I/O ports at %04lx\n", pos & PCI_BASE_ADDRESS_IO_MASK);
+	    {
+	      if (verbose > 1)
+		printf("\tRegion %d: ", i);
+	      else
+		putchar('\t');
+	      printf("I/O ports at %04lx\n", pos & PCI_BASE_ADDRESS_IO_MASK);
+	    }
 	}
       else if (cmd & PCI_COMMAND_MEMORY)
 	{
 	  int t = flg & PCI_BASE_ADDRESS_MEM_TYPE_MASK;
+	  if (verbose > 1)
+	    printf("\tRegion %d: ", i);
+	  else
+	    putchar('\t');
 	  printf("Memory at ");
 	  if (t == PCI_BASE_ADDRESS_MEM_TYPE_64)
 	    {
@@ -408,6 +425,11 @@ show_htype1(struct device *d)
 	   (brc & PCI_BRIDGE_CTL_FAST_BACK) ? '+' : '-');
 }
 
+static void
+show_htype2(struct device *d)
+{
+}
+
 static void
 show_verbose(struct device *d)
 {
@@ -421,34 +443,34 @@ show_verbose(struct device *d)
   byte max_lat, min_gnt;
   byte int_pin = get_conf_byte(d, PCI_INTERRUPT_PIN);
   byte int_line = get_conf_byte(d, PCI_INTERRUPT_LINE);
-  unsigned int irq, ex_htype;
+  unsigned int irq;
   word subsys_v, subsys_d;
 
   show_terse(d);
 
-  switch (class)
-    {
-    case PCI_CLASS_BRIDGE_PCI:
-      ex_htype = 1;
-      break;
-    default:
-      ex_htype = 0;
-    }
-  if (ex_htype != htype)
-    {
-      printf("\t!!! Header type %02x doesn't match class code %04x\n", htype, class);
-      return;
-    }
-
   switch (htype)
     {
-    case 0:
+    case PCI_HEADER_TYPE_NORMAL:
+      if (class == PCI_CLASS_BRIDGE_PCI)
+	{
+	badhdr:
+	  printf("\t!!! Header type %02x doesn't match class code %04x\n", htype, class);
+	  return;
+	}
       max_lat = get_conf_byte(d, PCI_MAX_LAT);
       min_gnt = get_conf_byte(d, PCI_MIN_GNT);
       subsys_v = get_conf_word(d, PCI_SUBSYSTEM_VENDOR_ID);
       subsys_d = get_conf_word(d, PCI_SUBSYSTEM_ID);
       break;
-    case 1:
+    case PCI_HEADER_TYPE_BRIDGE:
+      if (class != PCI_CLASS_BRIDGE_PCI)
+	goto badhdr;
+      irq = int_line = int_pin = min_gnt = max_lat = 0;
+      subsys_v = subsys_d = 0;
+      break;
+    case PCI_HEADER_TYPE_CARDBUS:
+      if ((class >> 8) != PCI_BASE_CLASS_BRIDGE)
+	goto badhdr;
       irq = int_line = int_pin = min_gnt = max_lat = 0;
       subsys_v = subsys_d = 0;
       break;
@@ -544,12 +566,15 @@ show_verbose(struct device *d)
 
   switch (htype)
     {
-    case 0:
+    case PCI_HEADER_TYPE_NORMAL:
       show_htype0(d);
       break;
-    case 1:
+    case PCI_HEADER_TYPE_BRIDGE:
       show_htype1(d);
       break;
+    case PCI_HEADER_TYPE_CARDBUS:
+      show_htype2(d);
+      break;
     }
 }
 
@@ -739,6 +764,8 @@ show_tree_dev(struct device *d, byte *line, byte *p)
         show_tree_bridge(b, line, p);
         return;
       }
+  if (verbose)
+    p += sprintf(p, "  %s", lookup_device_full(d->vendid, d->devid));
   print_it(line, p);
 }
 
@@ -844,6 +871,12 @@ main(int argc, char **argv)
       case 't':
 	show_tree++;
 	break;
+      case 'i':
+	pci_ids = optarg;
+	break;
+      case 'p':
+	pci_dir = optarg;
+	break;
       default:
       bad:
 	fprintf(stderr, help_msg);
